import os
import scanpy as sc

from ..logg import logger

dir_path = os.path.dirname(os.path.realpath(__file__)) + '/data/'
if not os.path.exists(dir_path): os.mkdir(dir_path)


def download_file(dir_download, web):
    if not os.path.exists(dir_download + '/neuron_10k_v3_filtered_feature_bc_matrix.h5'):
        logger.info('Downloading file with wget')
        os.system("wget -P {dir_download} {web}".format(dir_download=dir_download, web=web))


def preprocess_file(dir_download):
    import numpy as np
    import scipy as sp

    input_file = "{dir_download}/neuron_10k_v3_filtered_feature_bc_matrix.h5".format(dir_download=dir_download)
    logger.info("Preprocessing dataset using scanpy ({inf})".format(inf=input_file))

    adata = sc.read_10x_h5(input_file)

    adata.var_names_make_unique()
    adata.var_names = [name.upper() for name in adata.var_names.tolist()]
    sc.pp.filter_cells(adata, min_genes=20)
    sc.pp.filter_genes(adata, min_cells=10)

    X = adata.copy().X.astype(float)
    X[X > 100000] = 0  # Does not allow using np.isfinite()
    mean = np.mean(X, axis=0)
    X[X != 0] = 1
    per_zeros = np.sum(X, axis=0) / X.shape[1]

    # We will filter by mean, variance and percentage of zeros
    adata = adata[:, np.argwhere((mean > np.percentile(mean, 35)) |
                                 (per_zeros > np.percentile(per_zeros, 35))).ravel()]


    sc.write("{dir_download}/neurons10k.h5ad".format(dir_download=dir_download), adata)


def neurons10k():
    """
    Downloads 10x neurons dataset from E18 mouse.

    Parameters
    ----------

    Returns
    -------
    adata: :class:`scanpy.annData`
        AnnData object with with shape (11843 x 22240).
    """

    # Check final file
    neurons10kh5ad = dir_path + "/neurons10k.h5ad"

    try:
        adata = sc.read(neurons10kh5ad)

    except OSError:
        logger.info('10x brain 10k anndata file not found ({h5ad}). Downloading and processing file.'.format(h5ad=neurons10kh5ad))
        download_file(dir_path,
                      "http://cf.10xgenomics.com/samples/cell-exp/3.0.0/neuron_10k_v3/"
                      "neuron_10k_v3_filtered_feature_bc_matrix.h5")

        preprocess_file(dir_path)

        logger.info("Reading {} file".format(neurons10kh5ad))
        adata = sc.read(neurons10kh5ad)

    return adata



import sys
import inspect

def get_size(obj, seen=None):
    """Recursively finds size of objects in bytes"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if hasattr(obj, '__dict__'):
        for cls in obj.__class__.__mro__:
            if '__dict__' in cls.__dict__:
                d = cls.__dict__['__dict__']
                if inspect.isgetsetdescriptor(d) or inspect.ismemberdescriptor(d):
                    size += get_size(obj.__dict__, seen)
                break
    if isinstance(obj, dict):
        size += sum((get_size(v, seen) for v in obj.values()))
        size += sum((get_size(k, seen) for k in obj.keys()))
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum((get_size(i, seen) for i in obj))

    if hasattr(obj, '__slots__'): # can have __slots__ with __dict__
        size += sum(get_size(getattr(obj, s), seen) for s in obj.__slots__ if hasattr(obj, s))

    return size


############################################# UNITARY SCORE FUNCTIONS ##################################################

# Some of those functions may not be optimized with numba because they already work with sets.
# There are many similarity measures, some of the obtained from here
# http://users.uom.gr/~kouiruki/sung.pdf


def jaccard(A, B):
    return len(A & B)/len(A | B)


def wave_hedges(A, B):
    return (len(A - B) + len(B - A))/len(A | B)


def tanimoto(A, B):
    union = len(A | B)
    return (union - len(A & B))/union


def max_coefficient(A, B):
    if (len(A) == 0) | (len(B) == 0):
        return 0
    else:
        return len(A & B) / max(len(A), len(B))


def min_coefficient(A, B):
    if (len(A) == 0) | (len(B) == 0):
        return 0
    else:
        return len(A & B) / min(len(A), len(B))


def overlap_score(A, B):
    if (len(A) == 0) | (len(B) == 0):
        return 0
    else:
        return (len(A) + len(B)) / (2*len(A)*len(B)) * len(A & B)


def boolean(A, B, p):
    return 0 if len(A & B) / max(len(A), len(B)) < p else 1


def return_intersection_function(name='jaccard'):
    dict_funcs = {'JACCARD':jaccard, 'WAVE_HEDGES': wave_hedges, 'TANIMOTO': tanimoto,
                  'MAX': max_coefficient, 'MIN': min_coefficient, 'OVERLAP': overlap_score,
                  'BOOL': boolean}

    return dict_funcs[name.upper()]




from .joost2016 import joost2016
from .heart10k import heart10k
from .neurons10k import neurons10k

from ..logg import logger

import os

def delete_datasets(name = None):
    """
    Deletes all downloaded datasets.

    Parameters
    ----------
    name: ``str``
        Substring or string of dataset to remove ('joost2016').
    """
    dir_path = os.path.dirname(os.path.realpath(__file__)) + '/data'
    del_formats = ['h5ad', 'txt', 'mtx', 'csv', 'zip', 'tar', 'gz', 'rar', 'h5']

    for file in os.listdir(dir_path):
        for del_fmt in del_formats:
            if file.endswith(del_fmt):
                if name is None:
                    logger.info('Deleting {file}'.format(file=file))
                    os.remove(dir_path + '/' + file)
                else:
                    if file.__contains__(name):
                        logger.info('Deleting {file}'.format(file=file))
                        os.remove(dir_path + '/' + file)


from . import tl
from . import pl as pl
from . import pp as pp
from . import datasets as datasets
from .classes.scallop import Scallop
from .classes.bootstrap import Bootstrap

__author__ = ', '.join([
    'Alex M. Ascensión',
    'Olga Ibañez-Solé',
])
__email__ = ', '.join([
    'alexmascension@gmail.com',
    'olga.ibanez@biodonostia.org',
    # We don’t need all, the main authors are sufficient.
])

__version__ = "1.1.0"




import leidenalg
import louvain
from scanpy.utils import get_igraph_from_adjacency
import numpy as np


def do_clustering(alg_name, *args, **kwargs):
    supported_algnames = ['leiden', 'louvain']
    assert alg_name in supported_algnames, 'Supported algorithms are {}'.format(supported_algnames)

    if alg_name == 'leiden':
        return run_leiden(*args, **kwargs)
    elif alg_name == 'louvain':
        return run_louvain(*args, **kwargs)


def run_leiden(adj, resolution: float, random_state: int = None, use_weights: bool = True):
    # Remember that use_weights should be True to obtain a more realistic graph from the adjacency matrix.

    g = get_igraph_from_adjacency(adj, directed=True)

    weights = np.array(g.es['weight']).astype(np.float64) if use_weights else None
    part = leidenalg.find_partition(g, leidenalg.RBConfigurationVertexPartition, resolution_parameter=resolution,
                                    weights=weights, seed=random_state)
    groups = np.array(part.membership)
    return groups


def run_louvain(adj, resolution: float, random_state: int = None, use_weights: bool = True):
    g = get_igraph_from_adjacency(adj, directed=True)

    weights = np.array(g.es['weight']).astype(np.float64) if use_weights else None
    part = louvain.find_partition(g, louvain.RBConfigurationVertexPartition, resolution_parameter=resolution,
                                  weights=weights) # Todo: look if the update includes seed
    groups = np.array(part.membership)
    return groups




import numpy as np
import pandas as pd
from scipy import stats
# import networkx as nx # this is for conductance
import ray
from tqdm import tqdm

from . import Scallop
from ..tools._clustering import do_clustering, run_leiden
from ..tools._utils_bt import find_mapping
from ..tools._intersection_functions import return_intersection_function
from ..logg import logger
from ..tools._parallel_clustering import obtain_btmatrix_col
from ..tools._get_object_size import get_size


class Bootstrap():
    """
    Scallop object will store all Bootstraps experiments.

    Parameters
    ----------
    bt_id: ``int``
        ID of `Boostrap` object. This id is useful to retrieve the `Boostrap` object from
        `Scallop.list_bootstraps`.

    scal: :class:`scallop.Scallop`
        `scallop.Scallop` object

    res: ``float``
        Leiden resolution parameter.

    frac_cells: ``float``
        Proportion of cells that will be mapped. It can be in range [0-1].

    n_trials: ``int``
        Number of times bootstrapping will be repeated.
    """

    def __init__(self, bt_id: int,
                 scal: Scallop,
                 res: float,
                 frac_cells: float,
                 n_trials: int,
                 clustering: str):

        logger.debug('Bootstrap object created with parameters:\nid={}'
                     '\nres={}\nfrac_cells={}\nn_trials={}\nclustering={}'.format(
            bt_id, res, frac_cells, n_trials, clustering))

        self.scal = scal
        self.res = res
        self.frac_cells = frac_cells
        self.n_trials = n_trials
        self.sample_size = int(frac_cells * self.scal.n_cells)
        self.overlap_thresh = 0.1
        self.id = bt_id
        self.clustering = clustering

        self.empty_value = -1
        self.bt_matrix = np.full((self.scal.n_cells, self.n_trials), self.empty_value, dtype=int)
        self.mapped_matrix = None
        self.freq_score = None
        self.most_freq = None
        self.inter_score = None
        self.ident = None
        self.ident_clusters = None
        self.conductance = None

    def _strRepr(self):
        def IDformat(bt_id):
            """
            :param bt_id: Bootstrap ID
            :return: foratted Boostrap ID
            """
            id_str = str(bt_id)
            if len(id_str) == 1:
                id_str = ' ' + id_str
            return id_str

        line = '-' * 60 + '\n'
        string = 'Bootstrap ID: {} | res: {} | frac_cells: {} | n_trials: {} | clustering: {}'.format(
            IDformat(self.id), round(float(self.res), 1), self.frac_cells, self.n_trials, self.clustering)
        return line + string

    def __repr__(self):
        return self._strRepr()

    def __str__(self):
        return self._strRepr()

    def _getBtmatrix(self, n_procs=1):
        """
        Obtain the bootstrap matrix (self.bt_matrix), that is, a n_cells x n_trials matrix where, in each column, the
        cluster identities with the selected cells are shown. The cells that have not been selected will be filled with
        and exclusion value (self.empty_value) that by default is -1.
        """
        # --------------------------------------------------------------------------------------------------------------
        # Reference clustering (Leiden/Louvain clustering using all cells)
        adj = self.scal.annData.uns['neighbors']['connectivities']

        self.ident = do_clustering(self.clustering, adj=adj, resolution=self.res)
        self.ident_clusters = np.unique(self.ident)

        logger.info("Obtaining bootstrap matrix. {} processes are used at once.".format(n_procs))
        # --------------------------------------------------------------------------------------------------------------
        # Bootstrap iterations:
        if n_procs == 1:  # We keep this option because if the matrices are small or the n_trials is low, the
            # computation can be faster.

            for trial in tqdm(range(self.n_trials)):
                rnd_idx = np.random.choice(self.scal.n_cells, size=self.sample_size, replace=False)

                # We restrict the cells to the ones in rnd_idx
                adj_trial = adj[rnd_idx, :]
                adj_trial = adj_trial[:, rnd_idx]

                self.bt_matrix[rnd_idx, trial] = do_clustering(self.clustering, adj_trial, self.res)
        else:
            if ray.is_initialized():  # In case it remained opened from another process.
                ray.shutdown()

            ray.init(num_cpus=n_procs, memory=max(52428800, get_size(self.bt_matrix) * 10))

            list_clustering_arrs_ids = []

            adj_id = ray.put(adj)  # In this part we store the adjacency matrix in ray memory, so that all processes
            # use the same memory direction.

            '''
            In this part we apply ray. Ray requires the function to be called as function.remote(args), and the
            function calls are stored in a list (list_clustering_arrs_ids), as futures. Those futures are then executed
            with the method ray.get(list)
            '''
            for _ in range(self.n_trials):
                # TODO: add kwargs when they are available
                list_clustering_arrs_ids.append(obtain_btmatrix_col.remote(adj_id, self.scal.n_cells, self.sample_size,
                                                                       self.empty_value, self.clustering, self.res))

            list_clustering_arrs = ray.get(list_clustering_arrs_ids)
            ray.shutdown()

            self.bt_matrix = np.column_stack(list_clustering_arrs)

    def _cluster_mapping_mat(self, bt_col, method='overlap'):
        """
        Given two arrays with the labels of the original clustering, and the bootstrap clustering, returns the matrix
        where each (i,j) element is the similarity score between the cells that were
        assigned to cluster i in the reference clustering (Leiden clustering using all cells) and the cells that were
        assigned to cluster j in one of the bootstrap iterations. There are different similarity scores: 'overlap',
        'jaccard', etc.
        """

        bt_ident = self.bt_matrix[:, bt_col]
        clusters_bt = np.unique(bt_ident[bt_ident != self.empty_value])
        # If the number of clusters the two solutions (ident or bootstrap) are different, the matrix is filled with zeros
        # because munkres requires a square matrix
        n = max(len(self.ident_clusters), len(clusters_bt))
        mapping_matrix = np.zeros((n, n))
        for i in range(len(self.ident_clusters)):
            for j in range(len(clusters_bt)):
                cells_A = set(np.where(self.ident == self.ident_clusters[i])[0])
                cells_B = set(np.where(bt_ident == clusters_bt[j])[0])
                intersection_f = return_intersection_function(method)
                score = intersection_f(cells_A, cells_B)
                mapping_matrix[i][j] = score

        return mapping_matrix, clusters_bt

    def _renameIdent(self, method='overlap'):
        """
        Parameters
        ----------
        method: str
            Method of intersection score between two sets ('bool', 'overlap', 'jaccard', 'max', 'min').
        Returns
        -------
        None
        """
        # Todo: solve issue when freq(most freq) == freq(second most freq)
        # Initialize mapped_matrix with empty values. This matrix will contained the renamed cluster identities
        # (renamed: new cluster names that match those in reference solution)
        self.mapped_matrix = np.full((self.scal.n_cells, self.n_trials), self.empty_value)

        unk_count = len(self.ident_clusters)  # Name of the first unknown cluster (we start from the next integer after
        # known clusters

        for bt_trial in range(self.n_trials):

            # mapping_matrix: element (i, j) corresponds to the similarity score between cluster i from ident and
            # cluster j from bt_trial
            # clusters_bt_trial: list of unique cluster names in bootstrap solution
            mapping_matrix, clusters_bt_trial = self._cluster_mapping_mat(bt_col=bt_trial, method=method)

            # perm: permutation of the columns that maximizes the trace, in the form of a list of tuples
            # (a, b), where:
            #  a:  tuple index = original cluster name in in ident = position to which the column should be moved
            #  b: cluster name in bt trial = original position of the column that will be moved to a
            perm = find_mapping(mapping_matrix)

            # mapped_diag: diagonal of the permuted matrix
            mapped_diag = mapping_matrix[:, [item[1] for item in perm]].diagonal()

            # nonzero: positions of diag with value greater than overla_thresh. Why? Because we want to classify
            # mappings in 'good'/fixed clusters vs 'bad'/unknown clusters. Unknown mappings arise from
            # clusters that are more subdivided in a bootstrapping solution than in the original leiden solution.
            # Therefore, one of the parts 'takes' the cluster name (is renamed to the original cluster name) and the
            # other one is left as an unknown. If it matches an unknown from another trial, then they will both be
            # renamed to the same cluster number.
            nonzero = np.argwhere(mapped_diag > self.overlap_thresh).ravel()

            # nonzero_perm: permutation to be applied to non-zero positions
            nonzero_perm = [perm[i] for i in nonzero]

            # zero_perm: permutation to be applied to zero positions
            zero_perm = [perm[i] for i in np.argwhere(mapped_diag <= self.overlap_thresh).ravel()]

            # cluster_perm: dictionary with original cluster names in bt trial as keys and new names (corresponding name
            # from reference ident) as values
            clusters_perm = {clusters_bt_trial[tup[1]]: self.ident_clusters[tup[0]] for tup in nonzero_perm}

            # Deal with unknowns: clusters that were matched in the mapping matrix forcefully but that did not have a
            # decent overlap score will be 'tagged' by giving them names greater than len(unique(ident_clusters))
            for tup in zero_perm:
                if tup[1] in clusters_bt_trial:
                    clusters_perm[clusters_bt_trial[tup[1]]] = unk_count
                    unk_count += 1

            for c in list(clusters_perm.keys()):
                self.mapped_matrix[np.argwhere(self.bt_matrix[:, bt_trial] == c).ravel(), bt_trial] = clusters_perm[c]

    def _create_dict_unks(self):
        """
        :return: dictionary with keys: c cluster name, value: (col, idx) tuple where:
        col: column index of mapped_matrix where cluster c was found (integer)
        idx: list of row indices where c cluster was found within column col
        """
        dict_unks = {}
        unk_clus = len(self.ident_clusters)

        col = 0
        while unk_clus <= np.max(self.mapped_matrix):
            idx = np.argwhere(self.mapped_matrix[:, col] == unk_clus).ravel()
            if len(idx) > 0:
                dict_unks[unk_clus] = (col, idx)
                unk_clus += 1
            else:
                col += 1

        logger.debug('dict_unks: {}'.format(dict_unks))
        return dict_unks

    def _merge_equivalent_unks(self, dict_unks, threshold):
        """
        :param dict_unks: dictionary with {clus: (col, idx)} where:
        col: column index where clus is found within mapped_matrix
        idx: list of row indices where clus is found within col
        :param threshold: minimum similarity score two unknown clusters must have in order to be considered the same
        :return: dictionary with final cluster name: list of old cluster names that are considered to be the same
        """
        logger.info('Applying mapping of equivalent unknowns with threshold {}.'.format(threshold))
        unk_clus = len(self.ident_clusters)
        list_unk_clusters = np.sort(np.array(list(dict_unks.keys())))
        dict_merged_clusters = {}
        overlap_func = return_intersection_function('overlap')

        while len(list_unk_clusters) > 0:
            logger.debug('The list of clusters is {}'.format(list_unk_clusters))
            logger.debug('The list of columns is {}'.format([dict_unks[clust_i][0] for clust_i in list_unk_clusters]))

            clust_i = list_unk_clusters[0]
            list_overlap_vals = []

            for clust_j in list_unk_clusters:
                overlap_score = overlap_func(set(dict_unks[clust_i][1]), set(dict_unks[clust_j][1]))
                list_overlap_vals.append(overlap_score)
                logger.debug('Overlap threshold between {} (col {}, len {}) and {} (col {}, len {}) is {}.'.format(
                    clust_i, dict_unks[clust_i][0], clust_j, len(set(dict_unks[clust_i][1])),
                    dict_unks[clust_j][0], len(set(dict_unks[clust_j][1])), overlap_score))

            list_overlap_vals = np.array(list_overlap_vals)

            overlap_clus_idx = list_overlap_vals > threshold

            logger.debug('\n clust_i: {}\n list_overlap_vals: {} \n overlap_clus_idx: {}\n'
                         ''.format(clust_i, list_overlap_vals, overlap_clus_idx))

            for overlap_clus in list_unk_clusters[np.argwhere(overlap_clus_idx).ravel()]:
                dict_merged_clusters[overlap_clus] = unk_clus

            list_unk_clusters = list_unk_clusters[~ overlap_clus_idx]  # Select remaining positions
            unk_clus += 1

        logger.debug('dict_merged_clusters: {}'.format(dict_merged_clusters))
        return dict_merged_clusters

    def _remap_unks(self, threshold=0.9):
        """
        :param threshold: defined in method self._merge_equivalent_unks
        :return: updated mapped_matrix with remapped unknown clusters
        """
        dict_unks = self._create_dict_unks()
        dict_merged_clusters = self._merge_equivalent_unks(dict_unks=dict_unks, threshold=threshold)
        for original_clus, remapped_clus in dict_merged_clusters.items():
            col, idx = dict_unks[original_clus][0], dict_unks[original_clus][1]

            self.mapped_matrix[idx, col] = remapped_clus

    def _freqScore(self, do_return=False):
        """
        Obtains the frequency score for the sample, that is, the frequency of the most frequently assigned cluster
        identity per cell is stored in annData.obs['freq'].

        Parameters
        ----------
        do_return: bool
        If True, returns the freqScore pandas series.

        Returns
        -------
        return: pandas.Series
        Series object with freq score for each cell.
        """

        # Since, for each trial, a percentage of cells is empty, there is a probability that
        # in some cells the amount of empty values for all its trials is more than half of the
        # trials, or even all trials are empty. In the first case we create a nanMatrix in which
        # empty values are replaced by NaNs, so that the mode is not affected by empty values.
        # Then, the counts of the most frequent values are divided by the number of nonempty
        # trials for each cell. For the second case, since we have no information about the
        # clustering for that cell (all trials are empty values) we assign it a score of 0.
        # There might be other approaches, but this is not important right now, since we can
        # (and should) increase the number of trials in that case.
        nanmatrix = self.mapped_matrix.copy().astype(float)
        nanmatrix[nanmatrix == self.empty_value] = np.nan

        mode, count = stats.mode(nanmatrix, axis=1)

        del [nanmatrix]

        non_empty_values_count = np.sum((self.mapped_matrix != self.empty_value), axis=1)
        self.most_freq = mode.ravel()
        self.freq_score = count.ravel() / non_empty_values_count
        self.freq_score[self.freq_score == np.inf] = 0
        if do_return:
            return pd.DataFrame(self.freq_score,
                                index=self.scal.annData.obs_names, columns=['freqScore'])

    def _interScore(self):
        print('Doing Inter Score')
        pass



from munkres import Munkres
import numpy as np

def find_mapping(mapping_mat):
    """
    Applies Munkres algorithm for scoreMat to obtain the permutation that maximizes the trace.

    Parameters
    ----------
    scoreMat: numpy.ndarray
        Matrix where each (i,j) element is the score ('overlap', 'jaccard', etc.) between the cells in
        original bootstrap with cluster i, and cells in secondary bootstrap with cluster j.

    Returns
    -------
    perm: numpy.1darray
        Array with column permutation of scoreMat that maximizes the trace.
    """
    m = Munkres()
    perm = m.compute(1 - mapping_mat)
    # perm is a list of tuples (new_col, old_col)

    return perm




import logging

from ..classes.scallop import Scallop
from ..classes.bootstrap import Bootstrap
from ..logg import logger

import pandas as pd
import matplotlib.axes as ax
import ray


def getScore(scal: Scallop,
             res: [float, int] = 0.3,
             frac_cells: float = 0.95,
             n_trials: int = 25,
             score_type: str = "freq",
             clustering: str = "leiden",
             do_return: bool = False,
             n_procs: int = 1):
    """
    Obtains the score for each cell given resolution, clustering and the rest of parameters.

    Parameters
    ----------
    scal : :class:`Scallop`
        Scallop object

    res : ``int``,  ``float``
        Leiden resolution parameter.

    frac_cells : ``float``
        Proportion of cells that will be mapped. It can be in range [0-1].

    n_trials : ``int``
        Number of times bootstrapping will be repeated.

    score_type : ``str``
        Score type. Currently, only 'freq' is supported.

    clustering : ``str``
        Clustering performed in the dataset. Currently [``louvain``, ``leiden``] are allowed.

    do_return : ``bool``
        Return score as a ``pandas.Series`` object.

    n_procs : ``int``
        Number of processors for parallel execution.

    Returns
    -------
    score : :class:`pandas.Series`
        Object with score per cell, and the names of cells as index.
    """

    # We will first check if a Bootstrap object with the required parameters already exists.
    # If so, it will not generate a new Bootstrap object. Else, it appends a new Bootstrap
    # object to the Scallop object
    logger.info('Obtaining score with following parameters: \n '
                'Resolution: {res}\n Fraction of cells: {fc}\n Number of trials: {nt}\n Score type: {st}\n'
                'Clustering type: {ct}'.format(res=res, fc=frac_cells, nt=n_trials, st=score_type, ct=clustering))

    exists_bootstrap = False

    for idx in range(len(scal.list_bootstraps)):
        # Check whether a BT object exists with the same parameter values:
        if (scal.list_bootstraps[idx].res == res) and \
                (scal.list_bootstraps[idx].frac_cells == frac_cells) and \
                (scal.list_bootstraps[idx].n_trials == n_trials) and \
                (scal.list_bootstraps[idx].clustering == clustering):

            exists_bootstrap = True
            logger.info('A Bootstrap object with these parameter values already exists. '
                        'The score matrix will be returned directly.')
            bootstrap_obj = scal.list_bootstraps[idx]
            if do_return:
                return pd.DataFrame(scal.list_bootstraps[idx].freq_score, index=scal.annData.obs_names, columns=['freqScore'])
            break

    if not exists_bootstrap:
        idx = len(scal.list_bootstraps)
        # Because if the bootstrap does not exist, its ID will be equal to the length of list_bootstraps, as
        # the new botstrap will be added to the queue
        bootstrap_obj = Bootstrap(bt_id=len(scal.list_bootstraps),
                                  scal=scal,
                                  res=res,
                                  frac_cells=frac_cells,
                                  n_trials=n_trials,
                                  clustering=clustering)
        scal.list_bootstraps.append(bootstrap_obj)
        logger.info('The Bootstrap object did not exist and will been created.')

        logger.info('Calculating the bootstrap matrix.')
        bootstrap_obj._getBtmatrix(n_procs=n_procs)

    if score_type == 'freq':
        if bootstrap_obj.mapped_matrix is None:
            logger.info('Renaming identities.')
            bootstrap_obj._renameIdent()
            bootstrap_obj._remap_unks()

        if bootstrap_obj.freq_score is None:
            logger.info('Calculating freqScore.')
            bootstrap_obj._freqScore()

    elif score_type == 'inter':
        if bootstrap_obj.inter_score is None:
            bootstrap_obj._interScore()
    else:
        raise AttributeError('The score type can be "freq" or "inter".')

    if do_return:
        return pd.DataFrame(scal.list_bootstraps[idx].freq_score, index=scal.annData.obs_names, columns=['freqScore'])


def plotScore(scal: Scallop,
              score_type: str = "freq",
              plt_type: str = 'umap',
              bt_id: int = None,
              ax: ax.Axes = None,
              show: bool = True):
    """
    Plots the score from ``sl.tl.getScore()``

    Parameters
    ----------
    scal : :class:`Scallop`
        Scallop object

    score_type : ``str``
        Score type. Currently, only 'freq' is supported.

    plt_type : ``str``
        Plot type ("umap", "tsne", "phate", "pca").

    bt_id : ``int``
        Bootstrap id from scal.bootstrap_list. Run ``scal.getAllBootstraps()`` to see the
        id associated to certain conditions.

    ax : :class:`matplotlib.axes.Axes`
        Axis object in which store the plot.

    show : ``bool``
        Shows the plot on window.

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Stores the score axes into ``ax``.
    """

    scal._plotScore(score_type=score_type, plt_type=plt_type, bt_id=bt_id, ax=ax, show=show)



import numpy as np
import ray

from ._clustering import do_clustering

@ray.remote
def obtain_btmatrix_col(adj, n_cells, sample_size, empty_value, clustering, *args):
    rnd_idx = np.random.choice(n_cells, size=sample_size, replace=False)

    # We restrict the cells to the ones in rnd_idx
    adj_trial = adj[rnd_idx, :]
    adj_trial = adj_trial[:, rnd_idx]

    arr = np.full(n_cells, empty_value, dtype=int)
    # TODO: arr **kwargs when ray supports them
    arr[rnd_idx] = do_clustering(clustering, adj_trial, *args)
    return arr


import scanpy as sc
import scanpy.external as sce
import matplotlib.pyplot as plt
import numpy as np

from ..logg import logger

class Scallop():
    """
    Scallop object will store all Bootstraps experiments.

    Parameters
    ----------
    annData: class:`scanpy.annData`
        `scanpy.annData` object

    seed: ``int``
        Random seed for community detection.

    """

    """
    In previous versions, the object could be initialized from a path+filename to
    the gene expression matrix file as well as from a previously initialized annData
    object. For simplicity, we decided Scallop objects to be initialized
    from an annData object.
    """
    def __init__(self, annData: sc.AnnData, seed: int = 10):

        # Relative to the annData
        self.annData = annData
        self.n_cells = self.annData.shape[0]
        self.n_genes = self.annData.shape[1]
        self.cell_names = self.annData.obs.index.values
        self.gene_names = self.annData.var.index.values

        if 'neighbors' not in self.annData.uns:
            logger.warning('Neighbors not in annData. Computing neighbors')
            sc.pp.neighbors(self.annData)

        # Relative to other properties
        self.seed = seed

        # Relative to generated objects
        self.list_bootstraps = []

        # self.memory_use = optional parameter
    def __str__(self):
        return 'Scallop object with {} cells and {} genes'.format(self.n_cells, self.n_genes)

    def _plotScore(self,
                score_type: str = '',
                # we should add score_type once we implement interScore
                  plt_type: str = 'umap',
                  bt_id: int = 0, ax = None,
                  show: bool = True):
        # 'freq' or 'intersection'
        """
        Parameters
        ----------

        plt_type: str
            Plot type ("umap", "tsne", "phate", "pca").

        bt_id: int
            Bootstrap id from scal.bootstrap_list. Run `scal.getAllBootstraps()` to see the
            id associated to certain conditions.

        ax: matplotlib.Figure.Axis
            Axis object in which store the plot.

        show: bool
            Shows the plot on window.

        Returns
        -------
        Plots the results.
        """
        title = ''
        if ax is not None:
            show = False

        if plt_type == 'umap':
            if 'X_umap' not in self.annData.obsm.keys():
                sc.tl.umap(self.annData, random_state=self.seed)
            dr = self.annData.obsm['X_umap']
        elif plt_type == 'tsne':
            if 'X_tsne' not in self.annData.obsm.keys():
                sc.tl.tsne(self.annData, random_state=self.seed)
            dr = self.annData.obsm['X_tsne']
        elif plt_type == 'phate':
            if 'X_phate' not in self.annData.obsm.keys():
                sce.tl.phate(self.annData, random_state=self.seed)
            dr = self.annData.obsm['X_phate']
        elif plt_type == 'pca':
            if 'X_pca' not in self.annData.obsm.keys():
                sc.tl.pca(self.annData)
            dr = self.annData.obsm['X_pca']
        else:
            print("Enter a valid plot type ('umap', 'tsne', 'phate', 'pca'). Plotting umap.")
            dr = self.annData.obsm['umap']
            # Otherwise dr is referenced before assignment

        if bt_id is None:
            logger.warning('Bootstrap ID not provided. Last bootstrap will be plotted.')
            bt_id = len(self.list_bootstraps) - 1  # always plots last bootstrap

        n_plots = 3
        fig, axs = plt.subplots(nrows=1, ncols=n_plots, figsize=(10, 4))
        ident = self.list_bootstraps[bt_id].ident

        for c in np.unique(ident):
            idx = np.argwhere(ident == c).ravel()
            axs[0].scatter(dr[idx, 0], dr[idx, 1], s=2)
        most_freq = self.list_bootstraps[bt_id].most_freq

        for c in np.unique(most_freq):
            idx = np.argwhere(most_freq == c).ravel()
            axs[1].scatter(dr[idx, 0], dr[idx, 1], s=2)
            axs[2].set_title('freqScore')

        axs[2].scatter(dr[:, 0], dr[:, 1], c=self.list_bootstraps[bt_id].freq_score, s=2, cmap='Blues')
        plt.suptitle(title)

        # Todo: add colorbar
        axs[0].set_title('Identities in reference clustering')
        axs[1].set_title('Most frequently assigned cluster')
        axs[2].set_title('Frequency of assignment')
        for i in range(n_plots):
            axs[i].set_xticks([])
            axs[i].set_yticks([])

        plt.tight_layout()
        if show:
            plt.show()

    def getAllBootstraps(self, do_return: bool = False):
        """
        Prints a summary of each bootstrap element from the :class:`scallop.Scallop` object.

        Parameters
        ----------
        do_return : ``bool``
            Returns the printed string.

        Returns
        -------
        bt_str : ``str``
            String summary of :class:`scallop.Boostrap` objects.
        """
        bt_str = ''
        if len(self.list_bootstraps) == 0:
            bt_str += "No bootstraps found."
        else:
            for bt in self.list_bootstraps:
                bt_str += str(bt) + "\n"

        logger.info(bt_str)
        if do_return:
            return bt_str


